-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Top-level CipherCore for COLM_0
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use work.AEAD_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES      : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is
    signal en_chksum         : std_logic;
    signal sel_decrypth      : std_logic;
    signal sel_decryptl      : std_logic;
    signal sel_dec_final     : std_logic;
    signal sel_aesh          : std_logic_vector(2             -1 downto 0);
    signal sel_deltah        : std_logic_vector(3             -1 downto 0);
    signal sel_deltal        : std_logic_vector(3             -1 downto 0);
    signal sel_w             : std_logic;
    signal sel_tag           : std_logic;
    signal clr_w             : std_logic;
    signal en_l              : std_logic;
    signal en_deltah         : std_logic;
    signal en_deltal         : std_logic;
    signal en_w              : std_logic;
    signal en_aesh           : std_logic;
    signal en_aesl           : std_logic;
    signal en_tag            : std_logic;
    signal init_key          : std_logic;
        
    signal bdo_valid_s       : std_logic;
    signal done_init_key     : std_logic;
    signal aesh_almost_done  : std_logic;
    signal aesl_almost_done  : std_logic;
begin
    bdo_valid <= bdo_valid_s;
    
    u_cc_dp:
    entity work.CipherCore_Datapath(structure)
    generic map (G_MAX_LEN => G_MAX_LEN)
    port map (
        clk                => clk              ,
        rst                => rst              ,

        --! Input Processor
        key                => key              ,
        bdi                => bdi              ,
        bdi_valid_bytes    => bdi_valid_bytes  ,
        bdi_pad_loc        => bdi_pad_loc      ,
        bdi_size           => bdi_size         ,

        --! Output Processor
        bdo                => bdo              ,
        msg_auth_valid     => msg_auth_valid   ,

        --! Controller
        en_chksum          => en_chksum        ,
        sel_decrypth       => sel_decrypth     ,
        sel_decryptl       => sel_decryptl     ,
        sel_dec_final      => sel_dec_final    ,
        sel_aesh           => sel_aesh         ,
        sel_deltah         => sel_deltah       ,
        sel_deltal         => sel_deltal       ,
        sel_w              => sel_w            ,
        sel_tag            => sel_tag          ,
        clr_w              => clr_w            ,
        en_l               => en_l             ,
        en_deltah          => en_deltah        ,
        en_deltal          => en_deltal        ,
        en_w               => en_w             ,
        en_aesh            => en_aesh          ,
        en_aesl            => en_aesl          ,
        en_tag             => en_tag           ,
        en_bdo             => bdo_valid_s      ,
        init_key           => init_key         ,
        done_init_key      => done_init_key    ,
        aesh_almost_done   => aesh_almost_done  ,
        aesl_almost_done   => aesl_almost_done 
    );

    u_cc_ctrl:
    entity work.CipherCore_Control(behavior)
    generic map (G_MAX_LEN => G_MAX_LEN)
    port map (
        clk                => clk              ,
        rst                => rst              ,

        --! Input
        key_ready          => key_ready        ,
        key_valid          => key_valid        ,
        key_update         => key_update       ,
        mode               => decrypt          ,
        bdi_ready          => bdi_ready        ,
        bdi_valid          => bdi_valid        ,
        bdi_type           => bdi_type         ,
        bdi_eot            => bdi_eot          ,
        bdi_eoi            => bdi_eoi          ,
        bdi_size           => bdi_size         ,

        --! Datapath
        en_chksum          => en_chksum        ,
        sel_decrypth       => sel_decrypth     ,
        sel_decryptl       => sel_decryptl     ,
        sel_dec_final      => sel_dec_final    ,
        sel_aesh           => sel_aesh         ,
        sel_deltah         => sel_deltah       ,
        sel_deltal         => sel_deltal       ,
        sel_w              => sel_w            ,
        sel_tag            => sel_tag          ,
        clr_w              => clr_w            ,
        en_l               => en_l             ,
        en_deltah          => en_deltah        ,
        en_deltal          => en_deltal        ,
        en_w               => en_w             ,
        en_aesh            => en_aesh          ,
        en_aesl            => en_aesl          ,
        en_tag             => en_tag           ,
        init_key           => init_key         ,
        done_init_key      => done_init_key    ,
        aesh_almost_done   => aesh_almost_done  ,
        aesl_almost_done   => aesl_almost_done  ,
        --! Output
        msg_auth_done      => msg_auth_done    ,
        bdo_ready          => bdo_ready        ,
        bdo_valid          => bdo_valid_s
    );    
end structure;