-------------------------------------------------------------------------------
--! @file       CipherCore_Control.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Control unit for COLM_0
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity CipherCore_Control is
    generic (
        G_MAX_LEN           : integer := SINGLE_PASS_MAX
    );
    port (
        clk                 : in  std_logic;
        rst                 : in  std_logic;

        --! Input
        key_ready           : out std_logic;
        key_valid           : in  std_logic;
        key_update          : in  std_logic;

        mode                : in  std_logic;
        bdi_ready           : out std_logic;
        bdi_valid           : in  std_logic;
        bdi_type            : in  std_logic_vector(3             -1 downto 0);
        bdi_eot             : in  std_logic;
        bdi_eoi             : in  std_logic;
        bdi_size            : in  std_logic_vector(5             -1 downto 0);

        --! Control
        en_chksum           : out std_logic;
        sel_decrypth        : out std_logic;
        sel_decryptl        : out std_logic;
        sel_dec_final       : out std_logic;
        sel_aesh            : out std_logic_vector(2             -1 downto 0);
        sel_deltah          : out std_logic_vector(3             -1 downto 0);
        sel_deltal          : out std_logic_vector(3             -1 downto 0);
        sel_w               : out std_logic;
        sel_tag             : out std_logic;
        clr_w               : out std_logic;
        en_l                : out std_logic;
        en_deltah           : out std_logic;
        en_deltal           : out std_logic;
        en_w                : out std_logic;
        en_aesh             : out std_logic;
        en_aesl             : out std_logic;
        en_data             : out std_logic;
        en_tag              : out std_logic;
        init_key            : out std_logic;

        done_init_key       : in  std_logic;
        aesh_almost_done    : in  std_logic;
        aesl_almost_done    : in  std_logic;

        --! Output
        bdo_ready           : in  std_logic;
        bdo_valid           : out std_logic;
        msg_auth_done       : out std_logic
    );
end entity CipherCore_Control;

architecture behavior of CipherCore_Control is

    type tstate_type is (
        S_START,            S_LDKEY,
        S_INIT_L,           S_SAVE_L,
        S_PREP_DELTA_AD,
        S_NPUB,             S_AD,
        S_PREP_DELTA_D0,    S_PREP_DELTA_D1,    S_PREP_DELTA_D2,
        S_DATA,             S_PREP_FINAL_DECRYPT_WAIT_TAG,
        S_ENC_FINAL,        S_TAG_ENCRYPT,      S_TAG_DECRYPT,
        S_WAIT,             S_WAIT_DONE);
    signal tstate, ntstate      : tstate_type;
    signal nntstate, nntstate_r : tstate_type;

    type mstate_type is (
        S_INIT,
        S_DATA,             S_WAIT);
    signal mstate, nmstate      : mstate_type;

    type bstate_type is (
        S_INIT, S_INIT2, S_PREP_DELTAL_1, S_PREP_DELTAL_2,
        S_TAG_PT2_WAIT, S_DATA, S_TAG
    );
    signal bstate, nbstate      : bstate_type;

    signal bdi_full             : std_logic;
    signal aesh_valid           : std_logic;
    signal aesl_valid           : std_logic;
    signal aesl_busy            : std_logic;
    signal mid_ready            : std_logic;
    signal sel_decrypt_top      : std_logic;

    signal en_aeshh             : std_logic;
    signal en_aesll             : std_logic;
    signal en_deltal_s          : std_logic;
    signal en_ww                : std_logic;
    signal bdi_rdy              : std_logic;
    signal bdi_rdy2             : std_logic;
    signal bdo_vld              : std_logic;
    signal is_decrypt           : std_logic;
    signal is_aesh_proc         : std_logic;
    signal is_aesh_data         : std_logic;
    signal is_aesh_final        : std_logic;
    signal is_aesh_tag          : std_logic;
    signal is_aesh_empty        : std_logic;
    signal is_aesh_full         : std_logic;
    signal is_aesh_ignore       : std_logic;
    signal is_aesh_decrypt      : std_logic;
    signal is_aesl_data         : std_logic;
    signal is_aesl_final        : std_logic;
    signal is_aesl_tag          : std_logic;
    signal is_aesl_empty        : std_logic;
    signal is_aesl_full         : std_logic;

    signal bot_ready            : std_logic;

    signal msg_done             : std_logic;

begin
    bdi_ready <= bdi_rdy or bdi_rdy2;
    bdo_valid <= bdo_vld;
    bdi_full  <= bdi_size(4);
    en_chksum <= '1' when (is_decrypt = '0' and bdi_rdy = '1' and bdi_type(2 downto 1) = BDI_TYPE_DAT) or
                          (is_decrypt = '1' and bdi_rdy = '1' and bdi_type(2 downto 1) = BDI_TYPE_DAT and bdi_eot = '1')
                     else '0';

    p_reg:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                tstate      <= S_START;
                nntstate_r  <= S_START;
                mstate      <= S_INIT;
                bstate      <= S_INIT;

                aesh_valid <= '0';
            else
                tstate      <= ntstate;
                nntstate_r  <= nntstate;
                mstate      <= nmstate;
                bstate      <= nbstate;

                if (tstate = S_NPUB) then
                    is_decrypt <= mode;
                end if;


                --! Status registers for AES_H
                if (bdi_rdy = '1') then
                    is_aesh_proc <= '1';
                elsif (aesh_almost_done = '1') then
                    is_aesh_proc <= '0';
                end if;
                if (aesh_almost_done = '1' and is_aesh_ignore = '0') then
                    aesh_valid <= '1';
                elsif (en_ww = '1' or tstate = S_SAVE_L) then
                    aesh_valid <= '0';
                end if;
                if (tstate = S_START) then
                    is_aesh_data  <= '0';
                    is_aesh_empty <= '0';
                    is_aesh_full  <= '0';
                    is_aesh_ignore <= '0';
                elsif (bdi_rdy = '1') then
                    if (tstate = S_INIT_L) then
                        is_aesh_ignore <= '1';
                    else
                        is_aesh_ignore <= '0';
                    end if;
                    if (bdi_type(2 downto 1) = BDI_TYPE_DAT) then
                        is_aesh_data <= '1';
                    else
                        is_aesh_data <= '0';
                    end if;
                    if (bdi_size = 0) then
                        is_aesh_empty <= '1';
                    else
                        is_aesh_empty <= '0';
                    end if;                    
                    is_aesh_full    <= bdi_full;
                end if;
                if (en_aeshh = '1') then
                    is_aesh_decrypt <= sel_decrypt_top;
                end if;
                if (tstate = S_START) then
                    is_aesh_final <= '0';
                elsif (tstate = S_DATA and bdi_rdy = '1' and bdi_eot = '1') then
                    is_aesh_final <= '1';
                end if;
                if (tstate = S_START) then
                    is_aesh_tag <= '0';
                elsif ((tstate = S_TAG_ENCRYPT or tstate = S_TAG_DECRYPT) and en_aeshh = '1') then
                    is_aesh_tag <= '1';
                end if;

                --! Status registers for AES_L
                if (aesl_almost_done = '1') then
                    aesl_valid <= '1';
                elsif (tstate = S_START or en_deltal_s = '1') then
                    aesl_valid <= '0';
                end if;
                if (tstate = S_START or aesl_almost_done = '1') then
                    aesl_busy <= '0';
                elsif (en_aesll = '1') then
                    aesl_busy <= '1';
                end if;
                if (tstate = S_START) then
                    is_aesl_data   <= '0';
                    is_aesl_final  <= '0';
                    is_aesl_tag    <= '0';
                    is_aesl_empty  <= '0';
                    is_aesl_full   <= '0';
                elsif (en_aesll = '1') then
                    is_aesl_data   <= is_aesh_data ;
                    is_aesl_final  <= is_aesh_final;
                    is_aesl_tag    <= is_aesh_tag  ;
                    is_aesl_empty  <= is_aesh_empty;
                    is_aesl_full   <= is_aesh_full;
                end if;
            end if;
        end if;
    end process;

    en_aesh <= en_aeshh;
    en_aesl <= en_aesll;
    en_w    <= en_ww;
    en_deltal <= en_deltal_s;
    sel_tag <= msg_done;
    sel_decrypth <= sel_decrypt_top;
    sel_decryptl <= is_aesh_decrypt;

    p_state:
    process(tstate, bdi_valid, bdi_eot, bdi_type, key_valid, key_update, mid_ready, is_decrypt, aesh_almost_done,
            done_init_key, is_aesh_proc, msg_done, en_deltal_s, is_aesl_final, bdi_full, is_aesh_full)
    begin
        --! Default values
        ntstate          <= tstate;
        nntstate         <= nntstate_r;
        bdi_rdy         <= '0';
        en_l            <= '0';
        key_ready       <= '0';
        sel_aesh        <= "00";
        sel_deltah      <= "000";
        en_deltah       <= '0';
        en_aeshh        <= '0';
        en_data         <= '0';
        init_key        <= '0';
        clr_w           <= '0';
        sel_decrypt_top <= '0';

        case tstate is
            when S_START =>
                clr_w           <= '1';
                if (key_update = '0' and bdi_valid = '1') then
                    ntstate      <= S_PREP_DELTA_AD;
                elsif (key_update = '1' and key_valid = '1') then
                    ntstate      <= S_LDKEY;
                end if;

            when S_LDKEY =>
                key_ready <= '1';
                if (key_valid = '1') then
                    init_key    <= '1';
                    ntstate      <= S_WAIT;
                    nntstate     <= S_INIT_L;
                end if;

            when S_INIT_L =>
                sel_aesh <= "11";   --! Select 0
                en_aeshh <= '1';
                ntstate  <= S_WAIT;
                nntstate <= S_SAVE_L;

            when S_SAVE_L =>
                en_l        <= '1';
                ntstate     <= S_PREP_DELTA_AD;

            when S_PREP_DELTA_AD =>
                en_deltah   <= '1';
                sel_deltah  <= "111";        --! Delta = L
                ntstate     <= S_NPUB;

            when S_NPUB =>
                bdi_rdy     <= '1';
                sel_deltah  <= "001";        --! Delta = 3L = L1
                if (bdi_valid = '1') then
                    ntstate     <= S_WAIT;
                    nntstate    <= S_AD;
                    en_deltah   <= '1';
                    en_aeshh    <= '1';
                end if;

            when S_AD =>
                if (bdi_valid = '1' and mid_ready = '1') then
                    en_deltah   <= '1';
                    if (bdi_type(1) = '1') then     --! Not AD, init Block
                        sel_deltah  <= "111";
                        if (is_decrypt = '1') then
                            ntstate <= S_PREP_DELTA_D1;
                        else
                            ntstate <= S_DATA;
                        end if;
                    else
                        bdi_rdy     <= '1';
                        en_aeshh    <= '1';
                        if (bdi_eot = '1') then
                            if (bdi_full = '0') then
                                sel_deltah  <= "010";
                            end if;
                            ntstate     <= S_PREP_DELTA_D0;
                        else
                            ntstate     <= S_WAIT;
                        end if;
                    end if;
                end if;
                
            when S_PREP_DELTA_D0 =>
                en_deltah  <= '1';
                sel_deltah <= "111";
                nntstate   <= S_DATA;
                if (is_decrypt = '1') then
                    ntstate     <= S_PREP_DELTA_D1;
                elsif (is_aesh_proc = '1') then
                    ntstate     <= S_WAIT;
                else
                    ntstate     <= S_DATA;
                end if;

            when S_PREP_DELTA_D1 => --! x3
                en_deltah  <= '1';
                sel_deltah <= "001";
                ntstate    <= S_PREP_DELTA_D2;

            when S_PREP_DELTA_D2 => --! x3
                en_deltah <= '1';
                sel_deltah <= "001";                
                if (is_aesh_proc = '1') then
                    ntstate <= S_WAIT;
                else
                    ntstate <= S_DATA;
                end if;

            when S_DATA =>
                sel_decrypt_top <= is_decrypt;
                if (bdi_valid = '1' and mid_ready = '1') then
                    --! Perform operation except if it's the last block of decryption
                    --! If it's the last block of decryption, the input block must be concatenated with exp_tag to provide
                    --! the correct last block data.
                    if (bdi_eot = '1' and is_decrypt = '1' and bdi_full = '0') then
                        ntstate     <= S_PREP_FINAL_DECRYPT_WAIT_TAG;
                        sel_deltah  <= "011";    --! Delta = Delta (note: dont enable it, use it to control other signals instead)
                        bdi_rdy     <= '1';
                    else
                        if (bdi_eot = '0') then
                            en_deltah   <= '1';
                            en_aeshh    <= '1';                            
                            bdi_rdy     <= '1';
                            ntstate     <= S_WAIT;
                        elsif (bdi_eot = '1') then
                            if (bdi_full = '1') then
                                sel_deltah  <= "010";    --! x7
                            else
                                sel_deltah  <= "011";    --! x14
                            end if;
                            if (is_decrypt = '1') then
                                en_deltah   <= '1';
                                en_aeshh    <= '1';                            
                                bdi_rdy     <= '1';
                                ntstate     <= S_WAIT;
                                nntstate    <= S_TAG_DECRYPT;
                            else    --! For encryption, load the data onto checksum first
                                bdi_rdy  <= '1';    
                                ntstate     <= S_ENC_FINAL;                                
                            end if;
                        end if;
                    end if;
                end if;

            when S_PREP_FINAL_DECRYPT_WAIT_TAG =>
                --! Wait for the tag part of the data
                sel_aesh <= "01";
                sel_decrypt_top <= is_decrypt;
                if (bdi_valid = '1') then
                    sel_deltah  <= "011";    --! Delta = Delta x14
                    en_deltah   <= '1';
                    en_aeshh <= '1';
                    ntstate  <= S_WAIT;
                    nntstate <= S_TAG_DECRYPT;
                end if;

            when S_ENC_FINAL =>
                if (is_aesh_full = '1') then
                    sel_deltah  <= "010";    --! x7
                else
                    sel_deltah  <= "011";    --! x14
                end if;
                en_deltah   <= '1';
                sel_aesh    <= "11";        --! Get checksum
                en_aeshh    <= '1'; 
                ntstate <= S_WAIT;
                nntstate <= S_TAG_ENCRYPT;

            --! For tag (enc), take checksum
            when S_TAG_ENCRYPT =>
                sel_aesh <= "11";
                if (mid_ready = '1') then
                    en_aeshh    <= '1';
                    ntstate     <= S_WAIT;
                    nntstate    <= S_WAIT_DONE;
                end if;

            --! For decryption, we need to wait until the last block is decrypted
            when S_TAG_DECRYPT =>
                sel_aesh <= "10";
                sel_deltah <= "100";
                if (en_deltal_s = '1' and is_aesl_final = '1')  then
                    en_aeshh    <= '1';
                    ntstate     <= S_WAIT;
                    nntstate    <= S_WAIT_DONE;
                end if;

            when S_WAIT =>
                if (aesh_almost_done = '1' or done_init_key = '1') then
                    ntstate <= nntstate_r;
                end if;

            when S_WAIT_DONE =>
                if msg_done = '1' then
                    ntstate <= S_START;
                end if;


        end case;
    end process;


    p_mstate:
    process(mstate, tstate, is_decrypt, en_aeshh, bot_ready, aesl_busy, aesl_valid,
            aesh_valid, is_aesh_data, is_aesh_empty, is_aesl_tag, is_aesh_tag, aesl_almost_done, en_deltal_s)
    begin
        --! Default values
        nmstate         <= mstate;

        en_ww           <= '0';
        sel_w           <= '0';
        mid_ready       <= '0';
        en_aesll        <= '0';

        case mstate is
            when S_INIT =>
                if (tstate = S_NPUB and en_aeshh = '1') then
                    nmstate <= S_DATA;
                end if;

            when S_DATA =>
                if (aesl_busy = '0' and (aesl_valid = '0' or en_deltal_s = '1')) then
                    mid_ready  <= '1';
                end if;
                if (aesh_valid = '1' and aesl_busy = '0' and 
                        ((aesl_valid = '0') or
                         (bot_ready = '1') or 
                         (is_aesh_data = '0') or 
                         (is_aesh_empty = '1') or
                         (is_aesh_tag = '1' and is_decrypt = '1'))) 
                then
                    en_ww       <= '1';
                    if (is_aesh_data = '1') then
                        en_aesll    <= '1';
                        sel_w       <= not is_decrypt;
                        nmstate     <= S_WAIT;
                    end if;
                end if;

            when S_WAIT =>
                if (aesl_almost_done = '1') then
                    if (is_aesl_tag = '1') then
                        nmstate    <= S_INIT;
                    else
                        nmstate <= S_DATA;
                    end if;
                end if;
        end case;
    end process;



    p_bstate:
    process(bstate, tstate, aesl_valid, bdi_valid, bdo_ready, is_aesh_empty, is_aesl_final, is_decrypt, is_aesl_empty, is_aesl_full)
    begin
        --! Default values
        nbstate         <= bstate;

        sel_deltal      <= "000";
        en_deltal_s     <= '0';
        bot_ready       <= '0';
        bdo_vld         <= '0';
        msg_auth_done   <= '0';
        msg_done        <= '0';
        en_tag          <= '0';
        bdi_rdy2        <= '0';
        sel_dec_final   <= '0';

        case bstate is
            when S_INIT =>
                if (tstate = S_NPUB and bdi_valid = '1') then
                    nbstate <= S_INIT2;
                end if;

            when S_INIT2 => --! DletaL = L, w=0
                sel_deltal <= "111";
                if (is_decrypt = '1') then
                    nbstate <= S_DATA;
                else
                    nbstate <= S_PREP_DELTAL_1;
                end if;
                en_deltal_s <= '1';

            when S_PREP_DELTAL_1 => --! DeltaL = 3L
                en_deltal_s  <= '1';
                sel_deltal <= "001";
                nbstate    <= S_PREP_DELTAL_2;

            when S_PREP_DELTAL_2 => --! DeltaL = 3^2L
                en_deltal_s  <= '1';
                sel_deltal <= "001";
                nbstate    <= S_DATA;

            when S_DATA =>
                if (bdo_ready = '1' or (is_aesl_empty = '1' and aesl_valid = '1')) then
                    bot_ready <= '1';
                end if;
                if (aesl_valid = '1' and (bdo_ready = '1' or is_aesl_empty = '1')) then
                    en_deltal_s <= '1';
                    if (is_aesl_empty = '0') then
                        bdo_vld <= '1';
                    end if;                    
                    if (is_aesl_final = '1') then   --! DeltaL = DeltaL * 7
                        sel_dec_final <= is_decrypt;
                        en_tag     <= '1';
                        if (is_aesl_full = '1') then
                            sel_deltal <= "010";
                        else
                            sel_deltal <= "011";
                        end if;
                        nbstate <= S_TAG_PT2_WAIT;
                    end if;
                end if;

            when S_TAG_PT2_WAIT =>
                if (is_decrypt = '1') then
                    sel_deltal <= "100";
                end if;
                if (aesl_valid = '1') then
                    en_tag <= '1';
                    nbstate <= S_TAG;
                end if;

            when S_TAG =>
                if (is_decrypt = '1') then
                    sel_deltal <= "100";
                end if;
                if (aesl_valid = '1' and (bdo_ready = '1' or (is_decrypt = '1' and bdi_valid = '1'))) then
                    if (is_decrypt = '1') then
                        bdi_rdy2      <= '1';
                        msg_auth_done <= '1';
                    else
                        bdo_vld <= '1';
                    end if;
                    msg_done <= '1';
                    nbstate  <= S_INIT;
                end if;

        end case;
    end process;



end behavior;


