-------------------------------------------------------------------------------
--! @file       CipherCore_Datapath.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Datapath unit for COLM_0
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;
use work.AES_pkg.all;

entity CipherCore_Datapath is
    generic (
        G_SBOX_LOGIC        : boolean := False;
        G_MAX_LEN           : integer := SINGLE_PASS_MAX
    );
    port    (
        clk                 : in  std_logic;
        rst                 : in  std_logic;

        --! Input
        bdi                 : in  std_logic_vector(128           -1 downto 0);
        bdi_valid_bytes     : in  std_logic_vector(128/8         -1 downto 0);
        bdi_pad_loc         : in  std_logic_vector(128/8         -1 downto 0);
        key                 : in  std_logic_vector(128           -1 downto 0);
        bdi_size            : in  std_logic_vector(5             -1 downto 0);

        --! Control
        en_chksum           : in  std_logic;
        sel_decrypth        : in  std_logic;
        sel_decryptl        : in  std_logic;
        sel_dec_final       : in  std_logic;
        sel_aesh            : in  std_logic_vector(2             -1 downto 0);
        sel_deltah          : in  std_logic_vector(3             -1 downto 0);
        sel_deltal          : in  std_logic_vector(3             -1 downto 0);
        sel_w               : in  std_logic;
        sel_tag             : in  std_logic;
        clr_w               : in  std_logic;
        en_l                : in  std_logic;
        en_deltah           : in  std_logic;
        en_deltal           : in  std_logic;
        en_w                : in  std_logic;
        en_aesh             : in  std_logic;
        en_aesl             : in  std_logic;
        en_tag              : in  std_logic;
        en_bdo              : in  std_logic;
        init_key            : in  std_logic;

        done_init_key       : out std_logic;
        aesh_almost_done    : out std_logic;
        aesl_almost_done    : out std_logic;

        --! Output
        bdo                 : out std_logic_vector(128           -1 downto 0);
        msg_auth_valid      : out std_logic
    );
end entity CipherCore_Datapath;

architecture structure of CipherCore_Datapath is
    constant ZEROS          : std_logic_vector(128               -1 downto 0)
                            := (others => '0');

    --! GF(2^128) with x^128+x^7+x^2+x+1
    function mult2 (x: std_logic_vector(127 downto 0))
                   return std_logic_vector is
        variable xori : std_logic_vector(7 downto 0);
        variable xoro : std_logic_vector(7 downto 0);
    begin
        if x(127) = '1' then
            xori := x"87";
        else
            xori := x"00";
        end if;
        xoro := (x(6 downto 0) & '0') xor xori;

        return (x(126 downto 7) & xoro);
    end function mult2;

    function mult3 (x: std_logic_vector(127 downto 0))
                   return std_logic_vector is
    begin
        return (mult2(x) xor x);
    end function mult3;

    function mult7 (x: std_logic_vector(127 downto 0))
                   return std_logic_vector is
    begin
        return (mult3(mult2(x)) xor x);
    end function mult7;
    --! Registers
    signal ll               : std_logic_vector(128              -1 downto 0);
    signal ww               : std_logic_vector(128              -1 downto 0);

    signal deltah           : std_logic_vector(128              -1 downto 0);
    signal deltal           : std_logic_vector(128              -1 downto 0);
    signal tag              : std_logic_vector(128              -1 downto 0);
    signal vbytes_r         : std_logic_vector(128/8            -1 downto 0);
    signal pad_r            : std_logic_vector(128/8            -1 downto 0);
    signal zeros_r          : std_logic_vector(128/8            -1 downto 0);
    signal size_r           : std_logic_vector(5                -1 downto 0);
    --! Signals
    signal to_aeshxor       : std_logic_vector(128              -1 downto 0);
    signal to_aesh          : std_logic_vector(128              -1 downto 0);
    signal to_aesl          : std_logic_vector(128              -1 downto 0);
    signal to_wxor          : std_logic_vector(128              -1 downto 0);
    signal wwx3             : std_logic_vector(128              -1 downto 0);
    signal from_aesh        : std_logic_vector(128              -1 downto 0);
    signal from_aesl        : std_logic_vector(128              -1 downto 0);
    signal to_deltah        : std_logic_vector(128              -1 downto 0);
    signal to_deltal        : std_logic_vector(128              -1 downto 0);
    signal output           : std_logic_vector(128              -1 downto 0);
    signal out_sel          : std_logic_vector(128              -1 downto 0);
    signal tag_out          : std_logic_vector(128              -1 downto 0);
    signal chksum_sel       : std_logic_vector(128              -1 downto 0);
    signal outxor           : std_logic_vector(128              -1 downto 0);
            
    signal mult2h           : std_logic_vector(128              -1 downto 0);
    signal mult3h           : std_logic_vector(128              -1 downto 0);
    signal mult7h           : std_logic_vector(128              -1 downto 0);
    signal mult14h          : std_logic_vector(128              -1 downto 0);
    
    signal mult2l           : std_logic_vector(128              -1 downto 0);
    signal mult3l           : std_logic_vector(128              -1 downto 0);
    signal mult7l           : std_logic_vector(128              -1 downto 0);
    signal mult14l          : std_logic_vector(128              -1 downto 0);    
begin
    process(clk)
    begin
        if rising_edge(clk) then
            if (en_l = '1') then
                ll <= from_aesh;
            end if;
            --! Top
            if (clr_w = '1') then
                deltah <= (others => '0');
            elsif (en_deltah = '1') then
                deltah <= to_deltah;
            end if;
            --! Middle
            if (clr_w = '1') then
                ww <= (others => '0');
            elsif (en_w = '1') then
                ww <= to_wxor xor from_aesh;
            end if;
            --! Bottom
            if (en_deltal = '1') then
                deltal <= to_deltal;
            end if;
            --! Tag
            if (clr_w = '1') then
                tag <= (others => '0');
            elsif (en_tag = '1') then
                tag <= out_sel xor tag;
            end if;
            if (en_chksum = '1') then
                vbytes_r <= bdi_valid_bytes;
                pad_r    <= bdi_pad_loc;
                zeros_r  <= not (bdi_valid_bytes or bdi_pad_loc);
                size_r   <= bdi_size;                
            end if;            
        end if;
    end process;
    
    --! Top    
    topblock: block        
        -- signal bdi_bytes : std_logic_vector(128              -1 downto 0);
        signal rshf1     : std_logic_vector(128              -1 downto 0);
        signal rshf2     : std_logic_vector(128              -1 downto 0);
        signal rshf3     : std_logic_vector(128              -1 downto 0);
        signal rshf4     : std_logic_vector(128              -1 downto 0);
        signal data_r    : std_logic_vector(128              -1 downto 0);
        
        signal data_sel  : std_logic_vector(128              -1 downto 0);
        signal checksum  : std_logic_vector(128              -1 downto 0);
    begin
        process(clk)
        begin
            if rising_edge(clk) then
                if (clr_w = '1') then
                    data_r <= (others => '0');
                elsif (en_chksum = '1') then
                    data_r <= checksum;
                end if;
            end if;        
        end process;        
        checksum <= bdi xor data_r;
        
        gBytes: for i in 0 to 15 generate
            data_sel(7+i*8) <= data_r(7+i*8) xor rshf4(7+i*8) xor pad_r(i);
            data_sel(6+i*8 downto i*8) <= data_r(6+i*8 downto i*8) xor rshf4(6+i*8 downto i*8);
        end generate;
        
        rshf1 <= ZEROS( 7 downto 0) &   bdi(127 downto 8)  when size_r(0) = '1' else bdi;
        rshf2 <= ZEROS(15 downto 0) & rshf1(127 downto 16) when size_r(1) = '1' else rshf1;
        rshf3 <= ZEROS(31 downto 0) & rshf2(127 downto 32) when size_r(2) = '1' else rshf2;
        rshf4 <= ZEROS(63 downto 0) & rshf3(127 downto 64) when size_r(3) = '1' else rshf3;
        
        with sel_aesh select
        to_aeshxor <=   bdi                 when "00",
                        data_sel            when "01",
                        output              when "10",
                        data_r              when others;

        mult2h  <= mult2(deltah);
        mult3h  <= mult3(deltah);
        mult7h  <= mult2(mult3h) xor deltah;
        mult14h <= mult7(mult7h);

        with sel_deltah select
        to_deltah   <= mult2h    when "000",
                       mult3h    when "001",
                       mult7h    when "010",
                       mult14h   when "011",
                       mult2(to_deltal)    when "100",
                       ll        when others;
                       
        to_aesh     <= to_aeshxor xor to_deltah;
    end block;

    --! Middle
    to_wxor <= mult2(ww) when sel_w = '1' else ww;
    wwx3    <= mult3(ww);
    to_aesl <= wwx3 xor from_aesh;

    --! Bottom
    botblock: block
    begin
        mult2l  <= mult2(deltal);
        mult3l  <= mult3(deltal);
        mult7l  <= mult2(mult3l) xor deltal;
        mult14l <= mult7(mult7l);
        with sel_deltal select
        to_deltal   <= mult2l       when "000",
                       mult3l       when "001",
                       mult7l       when "010",
                       mult14l      when "011",
                       mult2h       when "100",
                       ll           when others;
        output <= from_aesl xor to_deltal;        
    end block;
    
    
    uAESh: entity work.AES_EncDec(structure)
    generic map (
        G_OBUF   => True
    )
    port map (
        clk             => clk              ,
        rst             => rst              ,
        din             => to_aesh          ,
        key             => key              ,
        dout            => from_aesh        ,
    
        init            => init_key         ,
        done_init       => done_init_key    ,
        decrypt         => sel_decrypth     ,
        start           => en_aesh          ,
        ready           => open             ,
        almost_done     => aesh_almost_done ,
        done            => open        
    );
    
    uAESL: entity work.AES_EncDec(structure)
    generic map (
        G_OBUF   => True
    )
    port map (
        clk             => clk              ,
        rst             => rst              ,
        din             => to_aesl          ,
        key             => key              ,
        dout            => from_aesl        ,
    
        init            => init_key         ,
        done_init       => open             ,
        decrypt         => sel_decryptl     ,
        start           => en_aesl          ,
        ready           => open             ,
        almost_done     => aesl_almost_done ,
        done            => open        
    );


    --! Output / Tag
    uBarrelrolatorL: block
        signal rol1, rol2, rol3, rol4 : std_logic_vector(128   -1 downto 0);
        signal chksum_out       : std_logic_vector(128         -1 downto 0);
    begin
        rol1 <= tag (119 downto 0) & tag (127 downto 120) when size_r(0) = '1' else tag;
        rol2 <= rol1(111 downto 0) & rol1(127 downto 112) when size_r(1) = '1' else rol1;
        rol3 <= rol2(95  downto 0) & rol2(127 downto 96) when size_r(2) = '1'  else rol2;
        rol4 <= rol3(63  downto 0) & rol3(127 downto 64) when size_r(3) = '1'  else rol3;
        tag_out <= rol4;
        
        process(clk)
        begin
            if rising_edge(clk) then
                if (clr_w = '1') then
                    chksum_out <= (others => '0');
                elsif (en_bdo = '1') then
                    chksum_out <= chksum_out xor output;
                end if;
            end if;
        end process;
        chksum_sel <= chksum_out when sel_dec_final = '1' else (others => '0');
    end block uBarrelrolatorL;
    
    outxor <= output xor chksum_sel;
    bdo <= tag_out when sel_tag = '1' else outxor;


    --! Compare
    cmpblock: block
        signal exp_tag                : std_logic_vector(128    -1 downto 0);
        signal zero_pass              : std_logic_vector(128/8  -1 downto 0);
        signal pad_pass               : std_logic_vector(128/8  -1 downto 0);
        signal vbytes                 : std_logic_vector(128/8  -1 downto 0);
        signal msg_auth_valid1        : std_logic;  
        signal msg_auth_valid1_r      : std_logic;  
        signal msg_auth_valid2        : std_logic;  
            
        signal ror1, ror2, ror3, ror4 : std_logic_vector(128    -1 downto 0);
    begin
        --! Check tha Pad location is x"80" and the subsequent bytes are zeros
        gCheck: for i in 0 to 15 generate
            zero_pass(i) <= '1' when zeros_r(i) = '0' or outxor(7+i*8 downto i*8) = x"00" else '0';
            pad_pass(i) <= '1' when pad_r(i) = '0' or outxor(7+i*8 downto i*8) = x"80" else '0';
        end generate;
        msg_auth_valid1 <= '1' when (zero_pass and pad_pass) = x"FFFF" else '0';

        process(clk)
        begin
            if rising_edge(clk) then
                if (sel_deltal(1) = '1') then
                    msg_auth_valid1_r <= msg_auth_valid1;
                end if;                                   
            end if;
        end process;

        --! Check for tag output (byte wise)
        vbytes <= not vbytes_r when sel_deltal(1) = '1' else vbytes_r;
        
        ror1 <= bdi (7 downto 0)   & bdi (127 downto 8)  when size_r(0) = '1' else bdi;
        ror2 <= ror1(15 downto 0)  & ror1(127 downto 16) when size_r(1) = '1' else ror1;
        ror3 <= ror2(31  downto 0) & ror2(127 downto 32) when size_r(2) = '1' else ror2;
        ror4 <= ror3(63  downto 0) & ror3(127 downto 64) when size_r(3) = '1' else ror3;
        
        gTagPt: for i in 0 to 15 generate
            out_sel(7+i*8 downto i*8) <= output(7+i*8 downto i*8) when vbytes(i) = '1' else (others => '0');
            exp_tag(7+i*8 downto i*8) <= ror4(7+i*8 downto i*8) when vbytes(i) = '1' else (others => '0');
        end generate;
        msg_auth_valid2 <= '1' when exp_tag = out_sel else '0';
        msg_auth_valid <= msg_auth_valid1_r and msg_auth_valid2;
    end block cmpblock;



end architecture structure;